#include <iostream>

// for std::pair
#include <utility>

// for std::sqrt
#include <cmath>

// for std::terminate
#include <exception>

#include <optional>

using namespace std;

struct SqrtError {};
struct QuadraticError {
  virtual string what() {
    return "generic quadratic eqn error";
  }
};
struct NonQuadraticEqn : public QuadraticError {
  string what() override {
    return "non-quadratic equation";
  }
};
struct NoRoots : public QuadraticError {
  string what() override {
    return "equation has no roots";
  }
};

// optional<int> is like int*
// o              |       p
// if (o) { ... } | if (p) { ... }
// o == nullopt   | p == nullptr
// *o             | *p

  

std::optional<char> FirstChar(const string& s) {
  if (s.size() == 0) {
    return std::nullopt;
  }
  return s[0];
}

// Compute square root using Newton's method
optional<double> Sqrt(double a) noexcept {
  if (a < 0) {
    return nullopt;
  }
  // calculate 2^(log₂ a) as the initial approximation
  double x = a;
  for (double t = a; t > 1; t /= 4, x /= 2) {}
  for (double t = a; t < 1; t *= 4, x *= 2) {}
  
  // 8 iterations are enough for doubles at this point
  for (int i = 0; i < 8; ++i) {
    // xᵢ₊₁ = xᵢ - f(xᵢ)/f'(xᵢ)
    //
    // where f(x) = x² - a, so the roots are +/- √a
    //       f'(x) = 2x
    x = (x + (a / x) ) / 2;
  }
  return x;
}

// auto x = Sqrt(5); // x = double OR x = an error value

// Find the roots of given quadratic equation
pair<double, double> SolveQuadraticEqn(double a, double b, double c) {
  if (a == 0) {
    throw NonQuadraticEqn{};
  }
  // square root of discriminant
  
  if (optional<double> r = Sqrt(b * b - 4 * a * c)) {
    double d = r.value();
    return { (-b + d) / (2 * a), (-b - d) / (2 * a) };
  } else {
    throw NoRoots{};
  }
}

void SolveAndPrint(double a, double b, double c) {
  try {
    auto [ root1, root2 ] = SolveQuadraticEqn(a, b, c);
    cout << "roots of " << a << "*x² + " << b << "*x + " << c << " = 0 are "
         << root1 << " and " << root2 << "\n";
  } catch (NonQuadraticEqn&) {
    if (b != 0) {
      auto root = - c / b;
      cout << "the root of " << a << "*x² + " << b << "*x + " << c << " = 0 is "
           << root << "\n";
    } else {
      cout << "eqn doesn't have roots\n";
    }
  } catch (NoRoots&) {
    cout << "could not solve the equation "
         << a << "*x² + " << b << "*x + " << c << " = 0\n";
  }
  cout << "solver done\n";
}

int main(int argc, char * argv[]) {
  // cout << "sqrt 2 = " << Sqrt(2) << "\n";
  // cout << "sqrt 4 = " << Sqrt(4) << "\n";
  // cout << "sqrt 1/2 = " << Sqrt(0.5) << "\n";
  // cout << "sqrt 3/2 = " << Sqrt(1.5) << "\n";
  // cout << "sqrt 1/4 = " << Sqrt(0.5) << "\n";
  // cout << "sqrt 1024 = " << Sqrt(1024) << "\n";
  // cout << "sqrt 2 = " << sqrt(2) << "\n";
  // cout << "sqrt 4 = " << sqrt(4) << "\n";
  // cout << "sqrt 1/2 = " << sqrt(0.5) << "\n";
  // cout << "sqrt 3/2 = " << sqrt(1.5) << "\n";
  // cout << "sqrt 1/4 = " << sqrt(0.5) << "\n";
  // cout << "sqrt 1024 = " << sqrt(1024) << "\n";
  // try {
  //   cout << "sqrt -3 = " << Sqrt(-3) << "\n";
  //   cout << "DONE\n";
  // } catch (SqrtError& s) {
  //   cout << "sqrt error\n";
  // }
  if (argc == 4) {
    double a = stod(argv[1]);
    double b = stod(argv[2]);
    double c = stod(argv[3]);
    try {
    auto [ root1, root2 ] = SolveQuadraticEqn(a, b, c);
    cout << "x1 = " << root1 << ", x2 = " << root2 << "\n";
    } catch (QuadraticError err) {
      cout << "could not solve the eqn: " << err.what() << "\n";
    }
  }

  cout << "example equations:\n";
  SolveAndPrint(1, 0, -4);
  SolveAndPrint(1, 4, 4);
  SolveAndPrint(1, 5, 6);
  SolveAndPrint(1, 0, 3);
  SolveAndPrint(1, -5, 6);
  SolveAndPrint(1, -5, 4);
  SolveAndPrint(1, 0, 4);
  SolveAndPrint(0, 1, 4);
  SolveAndPrint(0, 0, 4);

  return 0;
}
